// import 'source-map-support/register';
import { Context, Session } from 'koishi';
import { TabulatePluginConfig, TabulatePluginConfigLike } from './config';
import {
  DefinePlugin,
  InjectConfig,
  Inject,
  UseCommand,
  CommandUsage,
  CommandExample,
  PutSession,
  PutOption,
  PutArg,
  CommandDescription,
} from 'koishi-thirdeye';
import { Game, ReportScoreResult } from 'league-tabulator';
import { plainToInstance, Transform } from 'class-transformer';
import _ from 'lodash';
export * from './config';
import AragamiPlugin, { CacheKey } from 'koishi-plugin-cache-aragami';

class TabulateMatches {
  @CacheKey()
  location: string;

  @Transform((params) =>
    _.mapValues(params.value, (v) => plainToInstance(Game, v)),
  )
  games: Record<string, Game>;
}

@DefinePlugin({ name: 'tabulate', schema: TabulatePluginConfig })
export default class TabulatePlugin {
  constructor(private ctx: Context, config: TabulatePluginConfigLike) {}

  async getGames(key: string) {
    let data = await this.aragami.get(TabulateMatches, key);
    if (data) {
      return data;
    }
    data = new TabulateMatches();
    data.games = {};
    data.location = key;
    return data;
  }

  async saveGame(game: Game) {
    const matches = await this.getGames(game.location);
    matches.games[game.getKey()] = game;
    return this.aragami.set(matches, { ttl: this.config.saveTime });
  }

  async deleteGame(game: Game) {
    const matches = await this.getGames(game.location);
    if (!matches.games[game.getKey()]) {
      return;
    }
    delete matches.games[game.getKey()];
    return this.aragami.set(matches, { ttl: this.config.saveTime });
  }

  @InjectConfig()
  private config: TabulatePluginConfig;

  @Inject(true)
  private aragami: AragamiPlugin;

  @UseCommand('tabulate', '排表')
  @CommandDescription({ en: 'Tabulate' })
  @CommandUsage('创建一个新的比赛，并进行排表操作。')
  @CommandExample(
    'tabulate --kof --teama AA --teamb BB --playera "Nanahira,Momobako,Miko" --playerb "Yuzurisa,Senya,Maika" 进行一个 KOF 对局的排表',
  )
  async onTabulate(
    @PutSession() session: Session,
    @PutOption('location', '-p <location:string>  比赛地点，默认为群内。', {
      description: { en: 'Match location, defaults to current group.' },
    })
    location: string,
    @PutOption('rule', '-r <rule:string>  比赛模式', {
      fallback: '2/3【KOF】',
      description: { en: 'Match rule.' },
    })
    rule: string,
    @PutOption('kof', '--kof  使用 KOF 赛制', {
      description: { en: 'Use KOF rule.' },
    })
    useKof: boolean,
    @PutOption('head', '--head  使用 人头 赛制', {
      description: { en: 'Use single match rule.' },
    })
    useHead: boolean,
    @PutOption('teama', '--teama <teama:string>  A 队队名', {
      fallback: 'A',
      description: { en: 'Name for team A' },
    })
    teama: string,
    @PutOption('teamb', '--teamb <teamb:string>  B 队队名', {
      fallback: 'B',
      description: { en: 'Name for team B' },
    })
    teamb: string,
    @PutOption(
      'playera',
      '--playera <playera:string>  A 队队员名称，逗号分隔',
      {
        fallback: 'a1,a2,a3',
        description: {
          en: 'Names for members of team A, separated by commas.',
        },
      },
    )
    playera: string,
    @PutOption(
      'playerb',
      '--playerb <playera:string>  B 队队员名称，逗号分隔',
      {
        fallback: 'b1,b2,b3',
        description: {
          en: 'Names for members of team B, separated by commas.',
        },
      },
    )
    playerb: string,
  ) {
    const exactLocation =
      location || session.guildId || session.channelId || session.userId;
    if (!exactLocation) {
      return '无法获取比赛地点，请使用 --location 参数手动指定。';
    }
    const parsedPlayerA = playera.split(',');
    const parsedPlayerB = playerb.split(',');
    if (
      !parsedPlayerA.length ||
      !parsedPlayerB.length ||
      parsedPlayerA.length !== parsedPlayerB.length
    ) {
      return '两队人数不一致，无法排表。';
    }
    const game = new Game(
      exactLocation,
      new Date(),
      useKof ? '2/3【KOF】' : useHead ? '人头赛' : rule,
      teama,
      teamb,
      parsedPlayerA,
      parsedPlayerB,
    );
    await this.saveGame(game);
    return game.format();
  }

  @UseCommand('tabulate/score <self:integer> <oppo:integer>', '上报比分')
  @CommandDescription({ en: 'Report score for tabulated game' })
  @CommandUsage('进行比分的更新操作，并视情况进行下一轮或结束比赛。')
  @CommandExample('score -n Nanahira 2 0 标记当局中 Nanahira 以 2:0 战胜对方。')
  async updateScore(
    @PutSession() session: Session,
    @PutOption('location', '-p <location:string>  比赛地点，默认为群内。')
    location: string,
    @PutOption('name', '-n <name:string>  自己的 ID')
    name: string,
    @PutArg(0)
    self: number,
    @PutArg(1)
    oppo: number,
  ) {
    const exactLocation =
      location || session.guildId || session.channelId || session.userId;
    const possibleNames: string[] = _.uniq(
      _.compact([
        name,
        session.author?.nickname,
        session.author?.username,
        session.username,
        session.userId,
      ]),
    );
    const gamesMap = await this.getGames(exactLocation);
    const game = Object.values(gamesMap.games).find((game) =>
      game.includesPendingPlayer(possibleNames),
    );
    if (!game) {
      return '未找到比赛。请尝试使用 -n <name> 指定自己的名称。';
    }
    const matchingName = game.includesPendingPlayer(possibleNames);
    const result = game.reportScore(matchingName, self, oppo);
    if (result === ReportScoreResult.NotFound) {
      return `未找到对局。`;
    }
    if (result === ReportScoreResult.Finish) {
      await this.deleteGame(game);
    } else {
      await this.saveGame(game);
    }
    return game.format();
  }
}
